@@ -4,24 +4,25 @@
 This module provides low-level API for Tarantool
 '''
 
-import six
+import os
 import time
 import errno
+import socket
+
 import ctypes
 import ctypes.util
-import socket
-import msgpack
-
 try:
     from ctypes import c_ssize_t
 except ImportError:
     from ctypes import c_longlong as c_ssize_t
 
-import tarantool
+import msgpack
+
+import tarantool.error
 from tarantool.response import Response
 from tarantool.request import (
     Request,
-    RequestOK,
+    # RequestOK,
     RequestCall,
     RequestDelete,
     RequestEval,
@@ -32,28 +33,37 @@ from tarantool.request import (
     RequestSelect,
     RequestSubscribe,
     RequestUpdate,
-    RequestAuthenticate)
-
+    RequestUpsert,
+    RequestAuthenticate
+)
 from tarantool.space import Space
 from tarantool.const import (
     SOCKET_TIMEOUT,
     RECONNECT_MAX_ATTEMPTS,
     RECONNECT_DELAY,
-    RETRY_MAX_ATTEMPTS,
     REQUEST_TYPE_OK,
     REQUEST_TYPE_ERROR,
     IPROTO_GREETING_SIZE,
-    ENCODING_DEFAULT)
-
+    ITERATOR_EQ,
+    ITERATOR_ALL
+)
 from tarantool.error import (
     NetworkError,
     DatabaseError,
-    warn,
-    RetryWarning,
-    NetworkWarning)
-
-from .schema import Schema
-from .utils import check_key, greeting_decode, version_id
+    InterfaceError,
+    SchemaError,
+    NetworkWarning,
+    SchemaReloadException,
+    warn
+)
+from tarantool.schema import Schema
+from tarantool.utils import (
+    check_key,
+    greeting_decode,
+    version_id,
+    string_types,
+    ENCODING_DEFAULT,
+)
 
 
 class Connection(object):
@@ -66,10 +76,10 @@ class Connection(object):
     (insert/delete/update/select).
     '''
     Error = tarantool.error
-    DatabaseError = tarantool.error.DatabaseError
-    InterfaceError = tarantool.error.InterfaceError
-    SchemaError = tarantool.error.SchemaError
-    NetworkError = tarantool.error.NetworkError
+    DatabaseError = DatabaseError
+    InterfaceError = InterfaceError
+    SchemaError = SchemaError
+    NetworkError = NetworkError
 
     def __init__(self, host, port,
                  user=None,
@@ -78,7 +88,8 @@ class Connection(object):
                  reconnect_max_attempts=RECONNECT_MAX_ATTEMPTS,
                  reconnect_delay=RECONNECT_DELAY,
                  connect_now=True,
-                 encoding=ENCODING_DEFAULT):
+                 encoding=ENCODING_DEFAULT,
+                 call_16=False):
         '''
         Initialize a connection to the server.
 
@@ -88,11 +99,16 @@ class Connection(object):
         creates network connection.
                              if False than you have to call connect() manualy.
         '''
-        libc = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)
+        if os.name == 'nt':
+            libc = ctypes.WinDLL(
+                ctypes.util.find_library('Ws2_32'), use_last_error=True
+            )
+        else:
+            libc = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)
         recv = self._sys_recv = libc.recv
         recv.argtypes = [
             ctypes.c_int, ctypes.c_void_p, c_ssize_t, ctypes.c_int]
-        recv.restype = c_ssize_t
+        recv.restype = ctypes.c_int
         self.host = host
         self.port = port
         self.user = user
@@ -101,10 +117,13 @@ class Connection(object):
         self.reconnect_delay = reconnect_delay
         self.reconnect_max_attempts = reconnect_max_attempts
         self.schema = Schema(self)
+        self.schema_version = 1
         self._socket = None
         self.connected = False
+        self.inconnect = False
         self.error = True
         self.encoding = encoding
+        self.call_16 = call_16
         if connect_now:
             self.connect()
 
@@ -116,6 +135,12 @@ class Connection(object):
         self._socket = None
 
     def connect_basic(self):
+        if self.host == None:
+            self.connect_unix()
+        else:
+            self.connect_tcp()
+
+    def connect_tcp(self):
         '''
         Create connection to the host and port specified in __init__().
         :raise: `NetworkError`
@@ -127,7 +152,23 @@ class Connection(object):
             if self._socket:
                 self._socket.close()
             self._socket = socket.create_connection((self.host, self.port))
-            self._socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
+        except socket.error as e:
+            self.connected = False
+            raise NetworkError(e)
+
+    def connect_unix(self):
+        '''
+        Create connection to the host and port specified in __init__().
+        :raise: `NetworkError`
+        '''
+
+        try:
+            # If old socket already exists - close it and re-create
+            self.connected = True
+            if self._socket:
+                self._socket.close()
+            self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
+            self._socket.connect(self.port)
         except socket.error as e:
             self.connected = False
             raise NetworkError(e)
@@ -152,6 +193,7 @@ class Connection(object):
         :raise: `NetworkError`
         '''
         try:
+            self.inconnect = True
             self.connect_basic()
             self.handshake()
             # It is important to set socket timeout *after* connection.
@@ -159,8 +201,11 @@ class Connection(object):
             # the connection fails because the server is simply
             # not bound to port
             self._socket.settimeout(self.socket_timeout)
+            self.load_schema()
+            self.inconnect = False
         except socket.error as e:
             self.connected = False
+            self.inconnect = False
             raise NetworkError(e)
 
     def _recv(self, to_read):
@@ -168,13 +213,26 @@ class Connection(object):
         while to_read > 0:
             try:
                 tmp = self._socket.recv(to_read)
+            except OverflowError:
+                self._socket.close()
+                err = socket.error(
+                    errno.ECONNRESET,
+                    "Too big packet. Closing connection to server"
+                )
+                raise NetworkError(err)
             except socket.error:
-                raise NetworkError(socket.error(errno.ECONNRESET,
-                                   "Lost connection to server during query"))
+                err = socket.error(
+                    errno.ECONNRESET,
+                    "Lost connection to server during query"
+                )
+                raise NetworkError(err)
             else:
                 if len(tmp) == 0:
-                    raise NetworkError(socket.error(errno.ECONNRESET,
-                                       "Lost connection to server during query"))
+                    err = socket.error(
+                        errno.ECONNRESET,
+                        "Lost connection to server during query"
+                    )
+                    raise NetworkError(err)
                 to_read -= len(tmp)
                 buf += tmp
         return buf
@@ -199,24 +257,25 @@ class Connection(object):
         '''
         assert isinstance(request, Request)
 
-        # Repeat request in a loop if the server returns completion_status == 1
-        # (try again)
-        for attempt in range(RETRY_MAX_ATTEMPTS):    # pylint: disable=W0612
-            self._socket.sendall(bytes(request))
-            response = Response(self, self._read_response())
+        response = None
+        while True:
+            try:
+                self._socket.sendall(bytes(request))
+                response = Response(self, self._read_response())
+                break
+            except SchemaReloadException as e:
+                self.update_schema(e.schema_version)
+                continue
 
-            if response.completion_status != 1:
-                return response
-            warn(response.return_message, RetryWarning)
-
-        # Raise an error if the maximum number of attempts have been made
-        raise DatabaseError(response.return_code, response.return_message)
+        return response
 
     def _opt_reconnect(self):
         '''
         Check that connection is alive using low-level recv from libc(ctypes)
         **Due to bug in python - timeout is internal python construction.
         '''
+        if self.inconnect:
+            return
         if not self._socket:
             return self.connect()
 
@@ -228,25 +287,41 @@ class Connection(object):
                 if e.errno == errno.EBADF:
                     return errno.ECONNRESET
             else:
-                self._sys_recv(sock_fd, buf, 1,
-                               socket.MSG_DONTWAIT | socket.MSG_PEEK)
-                if ctypes.get_errno() == errno.EAGAIN:
+                if os.name == 'nt':
+                    flag = socket.MSG_PEEK
+                    self._socket.setblocking(False)
+                else:
+                    flag = socket.MSG_DONTWAIT | socket.MSG_PEEK
+                retbytes = self._sys_recv(sock_fd, buf, 1, flag)
+
+                err = 0
+                if os.name!= 'nt':
+                    err = ctypes.get_errno()
+                else:
+                    err = ctypes.get_last_error()
+                    self._socket.setblocking(True)
+
+
+                WWSAEWOULDBLOCK = 10035
+                if (retbytes < 0) and (err == errno.EAGAIN or
+                                       err == errno.EWOULDBLOCK or
+                                       err == WWSAEWOULDBLOCK):
                     ctypes.set_errno(0)
                     return errno.EAGAIN
-                return (ctypes.get_errno() if ctypes.get_errno()
-                        else errno.ECONNRESET)
+                else:
+                    return errno.ECONNRESET
 
         last_errno = check()
         if self.connected and last_errno == errno.EAGAIN:
             return
 
         attempt = 0
-        last_errno = 0
+        last_errno = errno.ECONNRESET
         while True:
             time.sleep(self.reconnect_delay)
             try:
                 self.connect_basic()
-            except NetworkError as e:
+            except NetworkError:
                 pass
             else:
                 if self.connected:
@@ -278,13 +353,20 @@ class Connection(object):
         assert isinstance(request, Request)
 
         self._opt_reconnect()
-        response = self._send_request_wo_reconnect(
-            request)
 
-        return response
+        return self._send_request_wo_reconnect(request)
+
+    def load_schema(self):
+        self.schema.fetch_space_all()
+        self.schema.fetch_index_all()
+
+    def update_schema(self, schema_version):
+        self.schema_version = schema_version
+        self.flush_schema()
 
     def flush_schema(self):
         self.schema.flush()
+        self.load_schema()
 
     def call(self, func_name, *args):
         '''
@@ -303,7 +385,7 @@ class Connection(object):
         if len(args) == 1 and isinstance(args[0], (list, tuple)):
             args = args[0]
 
-        request = RequestCall(self, func_name, args)
+        request = RequestCall(self, func_name, args, self.call_16)
         response = self._send_request(request)
         return response
 
@@ -331,7 +413,7 @@ class Connection(object):
     def replace(self, space_name, values):
         '''
         Execute REPLACE request.
-        It will throw error if there's no tuple with this PK exists
+        It won't throw error if there's no tuple with this PK exists
 
         :param int space_name: space id to insert a record
         :type space_name: int or str
@@ -341,12 +423,20 @@ class Connection(object):
 
         :rtype: `Response` instance
         '''
-        if isinstance(space_name, six.string_types):
+        if isinstance(space_name, string_types):
             space_name = self.schema.get_space(space_name).sid
         request = RequestReplace(self, space_name, values)
         return self._send_request(request)
 
     def authenticate(self, user, password):
+        '''
+        Execute AUTHENTICATE request.
+
+        :param string user: user to authenticate with
+        :param string password: password for the user
+
+        :rtype: `Response` instance
+        '''
         self.user = user
         self.password = password
         if not self._socket:
@@ -354,37 +444,63 @@ class Connection(object):
 
         request = RequestAuthenticate(self, self._salt, self.user,
                                       self.password)
-        return self._send_request_wo_reconnect(request)
+        auth_response = self._send_request_wo_reconnect(request)
+        if auth_response.return_code == 0:
+            self.flush_schema()
+        return auth_response
 
-    def join(self, server_uuid):
+    def _join_v16(self, server_uuid):
         request = RequestJoin(self, server_uuid)
-        sync = request._sync
-        resp = self._send_request(request)
+        self._socket.sendall(bytes(request))
+
         while True:
-            if self.version_id >= version_id(1, 7, 0):
-                # Send acknowledgement
-                ack = RequestOK(self, sync)
-                self._socket.sendall(bytes(ack))
+            resp = Response(self, self._read_response())
             yield resp
             if resp.code == REQUEST_TYPE_OK or resp.code >= REQUEST_TYPE_ERROR:
                 return
-            resp = Response(self, self._read_response())
         self.close()  # close connection after JOIN
 
-    def subscribe(self, cluster_uuid, server_uuid, vclock={}):
-        # FIXME rudnyh: ^ 'vclock={}'? really? sure?
-        request = RequestSubscribe(self, cluster_uuid, server_uuid, vclock)
-        sync = request._sync
-        resp = self._send_request(request)
+    def _join_v17(self, server_uuid):
+        class JoinState:
+            Handshake, Initial, Final, Done = range(4)
+
+        request = RequestJoin(self, server_uuid)
+        self._socket.sendall(bytes(request))
+        state = JoinState.Handshake
         while True:
-            if self.version_id >= version_id(1, 7, 0):
-                # Send acknowledgement
-                ack = RequestOK(self, sync)
-                self._socket.sendall(bytes(ack))
+            resp = Response(self, self._read_response())
             yield resp
             if resp.code >= REQUEST_TYPE_ERROR:
                 return
+            elif resp.code == REQUEST_TYPE_OK:
+                state = state + 1
+                if state == JoinState.Done:
+                    return
+
+    def _ops_process(self, space, update_ops):
+        new_ops = []
+        for op in update_ops:
+            if isinstance(op[1], string_types):
+                op = list(op)
+                op[1] = self.schema.get_field(space, op[1])['id']
+            new_ops.append(op)
+        return new_ops
+
+    def join(self, server_uuid):
+        self._opt_reconnect()
+        if self.version_id < version_id(1, 7, 0):
+            return self._join_v16(server_uuid)
+        return self._join_v17(server_uuid)
+
+    def subscribe(self, cluster_uuid, server_uuid, vclock=None):
+        vclock = vclock or {}
+        request = RequestSubscribe(self, cluster_uuid, server_uuid, vclock)
+        self._socket.sendall(bytes(request))
+        while True:
             resp = Response(self, self._read_response())
+            yield resp
+            if resp.code >= REQUEST_TYPE_ERROR:
+                return
         self.close()  # close connection after SUBSCRIBE
 
     def insert(self, space_name, values):
@@ -400,7 +516,7 @@ class Connection(object):
 
         :rtype: `Response` instance
         '''
-        if isinstance(space_name, six.string_types):
+        if isinstance(space_name, string_types):
             space_name = self.schema.get_space(space_name).sid
         request = RequestInsert(self, space_name, values)
         return self._send_request(request)
@@ -408,7 +524,8 @@ class Connection(object):
     def delete(self, space_name, key, **kwargs):
         '''
         Execute DELETE request.
-        Delete single record identified by `key` (using primary index).
+        Delete single record identified by `key`. If you're using secondary
+        index, it must be unique.
 
         :param space_name: space number or name to delete a record
         :type space_name: int or name
@@ -420,20 +537,122 @@ class Connection(object):
         index_name = kwargs.get("index", 0)
 
         key = check_key(key)
-        if isinstance(space_name, six.string_types):
+        if isinstance(space_name, string_types):
             space_name = self.schema.get_space(space_name).sid
-        if isinstance(index_name, six.string_types):
+        if isinstance(index_name, string_types):
             index_name = self.schema.get_index(space_name, index_name).iid
         request = RequestDelete(self, space_name, index_name, key)
         return self._send_request(request)
 
+    def upsert(self, space_name, tuple_value, op_list, **kwargs):
+        '''
+        Execute UPSERT request.
+
+        If there is an existing tuple which matches the key fields of
+        `tuple_value`, then the request has the same effect as UPDATE
+        and the [(field_1, symbol_1, arg_1), ...] parameter is used.
+
+        If there is no existing tuple which matches the key fields of
+        `tuple_value`, then the request has the same effect as INSERT
+        and the `tuple_value` parameter is used. However, unlike insert
+        or update, upsert will not read a tuple and perform error checks
+        before returning -- this is a design feature which enhances
+        throughput but requires more caution on the part of the user.
+
+        If you're using secondary index, it must be unique.
+
+        List of operations allows to update individual fields.
+
+        *Allowed operations:*
+
+        (For every operation you must provide field number, to apply this
+        operation to)
+
+        * `+` for addition (values must be numeric)
+        * `-` for subtraction (values must be numeric)
+        * `&` for bitwise AND (values must be unsigned numeric)
+        * `|` for bitwise OR (values must be unsigned numeric)
+        * `^` for bitwise XOR (values must be unsigned numeric)
+        * `:` for string splice (you must provide `offset`, `count` and `value`
+          for this operation)
+        * `!` for insertion (provide any element to insert)
+        * `=` for assignment (provide any element to assign)
+        * `#` for deletion (provide count of fields to delete)
+
+        :param space_name: space number or name to update a record
+        :type space_name: int or str
+        :param index: index number or name to update a record
+        :type index: int or str
+        :param tuple_value: tuple, that
+        :type tuple_value:
+        :param op_list: list of operations. Each operation
+            is tuple of three (or more) values
+        :type op_list: a list of the form [(symbol_1, field_1, arg_1),
+            (symbol_2, field_2, arg_2_1, arg_2_2, arg_2_3),...]
+
+        :rtype: `Response` instance
+
+        Operation examples:
+
+        .. code-block:: python
+
+            # 'ADD' 55 to second field
+            # Assign 'x' to third field
+            [('+', 2, 55), ('=', 3, 'x')]
+            # 'OR' third field with '1'
+            # Cut three symbols starting from second and replace them with '!!'
+            # Insert 'hello, world' field before fifth element of tuple
+            [('|', 3, 1), (':', 2, 2, 3, '!!'), ('!', 5, 'hello, world')]
+            # Delete two fields starting with second field
+            [('#', 2, 2)]
+        '''
+        index_name = kwargs.get("index", 0)
+
+        if isinstance(space_name, string_types):
+            space_name = self.schema.get_space(space_name).sid
+        if isinstance(index_name, string_types):
+            index_name = self.schema.get_index(space_name, index_name).iid
+        op_list = self._ops_process(space_name, op_list)
+        request = RequestUpsert(self, space_name, index_name, tuple_value,
+                                op_list)
+        return self._send_request(request)
+
     def update(self, space_name, key, op_list, **kwargs):
         '''
         Execute UPDATE request.
-        Update single record identified by `key` (using primary index).
+
+        The `update` function supports operations on fields — assignment,
+        arithmetic (if the field is unsigned numeric), cutting and pasting
+        fragments of a field, deleting or inserting a field. Multiple
+        operations can be combined in a single update request, and in this
+        case they are performed atomically and sequentially. Each operation
+        requires specification of a field number. When multiple operations are
+        present, the field number for each operation is assumed to be relative
+        to the most recent state of the tuple, that is, as if all previous
+        operations in a multi-operation update have already been applied.
+        In other words, it is always safe to merge multiple update invocations
+        into a single invocation, with no change in semantics.
+
+        Update single record identified by `key`.
 
         List of operations allows to update individual fields.
 
+        *Allowed operations:*
+
+        (For every operation you must provide field number, to apply this
+        operation to)
+
+        * `+` for addition (values must be numeric)
+        * `-` for subtraction (values must be numeric)
+        * `&` for bitwise AND (values must be unsigned numeric)
+        * `|` for bitwise OR (values must be unsigned numeric)
+        * `^` for bitwise XOR (values must be unsigned numeric)
+        * `:` for string splice (you must provide `offset`, `count` and `value`
+          for this operation)
+        * `!` for insertion (before) (provide any element to insert)
+        * `=` for assignment (provide any element to assign)
+        * `#` for deletion (provide count of fields to delete)
+
         :param space_name: space number or name to update a record
         :type space_name: int or str
         :param index: index number or name to update a record
@@ -441,19 +660,34 @@ class Connection(object):
         :param key: key that identifies a record
         :type key: int or str
         :param op_list: list of operations. Each operation
-            is tuple of three values
-        :type op_list: a list of the form
-            [(field_1, symbol_1, arg_1), (field_2, symbol_2, arg_2),...]
+            is tuple of three (or more) values
+        :type op_list: a list of the form [(symbol_1, field_1, arg_1),
+            (symbol_2, field_2, arg_2_1, arg_2_2, arg_2_3), ...]
 
-        :rtype: `Response` instance
+        :rtype: ``Response`` instance
+
+        Operation examples:
+
+        .. code-block:: python
+
+            # 'ADD' 55 to second field
+            # Assign 'x' to third field
+            [('+', 2, 55), ('=', 3, 'x')]
+            # 'OR' third field with '1'
+            # Cut three symbols starting from second and replace them with '!!'
+            # Insert 'hello, world' field before fifth element of tuple
+            [('|', 3, 1), (':', 2, 2, 3, '!!'), ('!', 5, 'hello, world')]
+            # Delete two fields starting with second field
+            [('#', 2, 2)]
         '''
         index_name = kwargs.get("index", 0)
 
         key = check_key(key)
-        if isinstance(space_name, six.string_types):
+        if isinstance(space_name, string_types):
             space_name = self.schema.get_space(space_name).sid
-        if isinstance(index_name, six.string_types):
+        if isinstance(index_name, string_types):
             index_name = self.schema.get_index(space_name, index_name).iid
+        op_list = self._ops_process(space_name, op_list)
         request = RequestUpdate(self, space_name, index_name, key, op_list)
         return self._send_request(request)
 
@@ -520,15 +754,21 @@ class Connection(object):
         offset = kwargs.get("offset", 0)
         limit = kwargs.get("limit", 0xffffffff)
         index_name = kwargs.get("index", 0)
-        iterator_type = kwargs.get("iterator", 0)
+        iterator_type = kwargs.get("iterator")
+
+        if iterator_type is None:
+            iterator_type = ITERATOR_EQ
+            if key is None or (isinstance(key, (list, tuple)) and
+                               len(key) == 0):
+                iterator_type = ITERATOR_ALL
 
         # Perform smart type checking (scalar / list of scalars / list of
         # tuples)
         key = check_key(key, select=True)
 
-        if isinstance(space_name, six.string_types):
+        if isinstance(space_name, string_types):
             space_name = self.schema.get_space(space_name).sid
-        if isinstance(index_name, six.string_types):
+        if isinstance(index_name, string_types):
             index_name = self.schema.get_index(space_name, index_name).iid
         request = RequestSelect(self, space_name, index_name, key, offset,
                                 limit, iterator_type)
@@ -550,7 +790,7 @@ class Connection(object):
         return Space(self, space_name)
 
     def generate_sync(self):
-        """\
+        '''
         Need override for async io connection
-        """
+        '''
         return 0
